
/*
Exemplo cliente-servidor
Codigo do servidor
(C) 2011-2012 Pedro Freire

Compilar (em Mac/Unix) com
	gcc -Wall server.c sha1.c -o server
ou em Windows com
	gcc -Wall server.c sha1.c -o server -lws2_32

Correr com
	server <porta-do-servidor>
ou
	server
(esta ultima cria um servidor com um numero de porta automatico)

E' possivel falar com este servidor usando uma ligacao Telnet normal.

Atencao que e' importante permitir que a firewall do sistema operativo
crie uma excepcao para o servidor, senao ele nao vai receber ligacoes!
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <assert.h>

#include "polarssl/sha1.h"

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#ifdef errno
#undef errno
#endif
#define EADDRNOTAVAIL		WSAEADDRNOTAVAIL
#define EADDRINUSE		WSAEADDRINUSE
#define ETIMEDOUT		WSAETIMEDOUT
#define ECONNREFUSED		WSAECONNREFUSED
#define ECONNRESET		WSAECONNRESET
#define perror_and_exit(msg)	{ fprintf(stderr, "Erro %i do %s do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms740668(v=vs.85).aspx\npara saber a descricao.\n", (int) WSAGetLastError(), msg); exit(EXIT_FAILURE); }
#define errno			WSAGetLastError()
#define sleep(s)		Sleep((s)*1000)
typedef int socklen_t;
#define IPPORT_USERRESERVED	5000
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#define INVALID_SOCKET		-1
#define SOCKET_ERROR		-1
#define perror_and_exit(msg)	{ perror(msg); exit(EXIT_FAILURE); }
#endif


#define DEFAULT_PORT	IPPORT_USERRESERVED

#define TEXT_WELCOME	"Benvindo ao servidor!\n"

#define TEXT_HELP	"\n"								\
			"Para eu repetir uma FRASE, escreva:\n"				\
			"\tdiz FRASE\n"							\
			"Para eu calcular uma raiz quadrada de um NUMERO, escreva:\n"	\
			"\traiz NUMERO\n"						\
			"Para eu calcular o valor de difusao SHA-1 de um ficheiro, escreva:\n"	\
			"\tsha1 TAMANHO-DO-FICHEIRO\n"					\
			"\tCONTEUDO-DO-FICHEIRO\n"					\
			"Para eu desligar, escreva:\n"					\
			"\tdesliga\n"							\
			"Para ajuda, usa qualquer outra frase!\n"			\
			"\n"


/* Le^ um ficheiro de "size" bytes a partir do socket "client_socket".
   Devolve um apontador para um buffer que e' alocado com malloc(), ou
   NULL em caso de erro/EOF.
   E' responsabilidade do chamador fazer free() a esse apontador!
   Usa leftover que aponta para leftover_size bytes de dados, como dados
   iniciais para o ficheiro (e.g.: lidos juntamente com o comando "sha1").
*/
unsigned char *recv_file( int client_socket, int size, const char *leftover, int leftover_size )
{
	unsigned char  *file_buffer, *p;
	int            recv_sz;

	file_buffer = malloc( size+1 );  /* +1 para debug (ver abaixo) */
	if( file_buffer == NULL )
		return NULL;  /* Nao consegue reservar memoria para o fich */

	p = file_buffer;
	if( leftover_size > 0 )
		{
		assert( leftover_size <= size );
		assert( leftover != NULL );
		memcpy( file_buffer, leftover, leftover_size );
		p    += leftover_size;
		size -= leftover_size;
		}
	while( size > 0 )
		{
		/* Le^ a linha enviada pelo cliente
		*/
		recv_sz = recv( client_socket, p, size, 0 );
		if( recv_sz == SOCKET_ERROR )
			{
			if( errno != EPIPE  &&  errno != ECONNRESET )
				perror_and_exit( "recv()" );  /* houve um erro, ou... */
			/* ...o outro lado desligou-se */
			recv_sz = 0;
			}
		if( recv_sz == 0 )
			{
			free( file_buffer );
			return NULL;  /* EOF = o cliente desligou */
			}
		assert( recv_sz <= size );
		p    += recv_sz;
		size -= recv_sz;
		}

	/* Para debug:
		*p = '\0';
		printf( "Recebi: \"%s\"\n", file_buffer );
	*/
	return file_buffer;
}


int main( int argc, char *argv[] )
{
	uint16_t            port;
	char                buffer[1024+1], buffer2[1024+1];
	unsigned char       *file_buffer;
	double              num;
	int                 size;
	int                 server_socket, client_socket;
	struct sockaddr_in  server_sa;  /* sockaddr deste servidor */
	struct sockaddr_in  client_sa;  /* sockaddr do cliente */
	socklen_t           client_sa_sz;
	int                 recv_sz;
	int                 i;
	char                *p;

#ifdef WIN32
	WSADATA wsaData;
	int wsaErr = WSAStartup( MAKEWORD(2,2), &wsaData );
	if( wsaErr != 0 )
		{
		fprintf( stderr, "Erro %i do WSAStartup() do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms742213(v=vs.85).aspx\npara saber a descricao.\n", (int) wsaErr );
		exit( EXIT_FAILURE );
		}
#else
	signal( SIGPIPE, SIG_IGN );
#endif

	/* Establecer a porta IP para o servidor
	*/
	port = (argc > 1 ? (uint16_t) atol(argv[1]) : DEFAULT_PORT );
	if( port < IPPORT_USERRESERVED )
		{
		fprintf( stderr, "Porta incorrecta (%u). Deveria ser >= a %u para evitar conflitos.\n", port, IPPORT_USERRESERVED );
		exit( EXIT_FAILURE );
		}

	/* Obter um socket (uma "tomada") para falar IPv4
	*/
	server_socket = socket( PF_INET, SOCK_STREAM, 0 );
	if( server_socket == INVALID_SOCKET )
		perror_and_exit( "socket()" );

	/* Atribuir (bind) o endereco IPv4 deste servidor a este socket;
	   Com isto o socket passa a estar acessivel externamente
	*/
	memset( &server_sa, 0, sizeof(server_sa) );
	server_sa.sin_family = AF_INET;
	server_sa.sin_addr.s_addr = htonl( INADDR_ANY );
	server_sa.sin_port = htons( port );
	if( bind(server_socket, (struct sockaddr*) &server_sa, sizeof(server_sa)) == SOCKET_ERROR )
		perror_and_exit( "bind()" );

	/* Permitir que os clientes se liguem (1 de cada vez) a este socket;
	   Atencao que isto so' funcionara' se for aberta esta porta na
	   firewall do sistema!
	*/
	if( listen(server_socket, 1) == SOCKET_ERROR )
		perror_and_exit( "listen()" );

	/* Informar o utilizador de onde esta' este servidor ligado
	*/
	if( server_sa.sin_family == AF_INET )
		{
		printf( "Estou 'a espera de ligacoes na porta IPv4 %u...\n\n",
			ntohs(server_sa.sin_port) );
			/* o IP e' exibido como "0.0.0.0" */
		}
	else
		puts( "Estou 'a espera de ligacoes numa familia de enderecos que nao IPv4...\n" );


	for(;;)
		{
		/* Esperar que um cliente se ligue
		*/
		client_sa_sz = sizeof( client_sa );
		client_socket = accept( server_socket, (struct sockaddr*) &client_sa, &client_sa_sz );
		if( client_socket == INVALID_SOCKET )
			perror_and_exit( "accept()" );

		/* Exibir informacao sobre quem se ligou
		*/
		if( client_sa.sin_family == AF_INET )
			{
			printf( "Ligou-se um cliente a partir do endereco IPv4 %s:%u\n",
				inet_ntoa(client_sa.sin_addr), ntohs(client_sa.sin_port) );
			}
		else
			puts( "Ligou-se um cliente a partir de uma familia de enderecos que nao IPv4" );

		/* Envia mensagem de boas-vindas e ajuda
		   e se nao houver erro, executa o ciclo principal
		*/
		if( send(client_socket, TEXT_WELCOME TEXT_HELP, strlen(TEXT_WELCOME TEXT_HELP), 0) == INVALID_SOCKET )
			{
			if( errno != EPIPE  &&  errno != ECONNRESET )
				perror_and_exit( "send()" );  /* houve um erro, ou... */
			/* ...o outro lado desligou-se */
			}
		else for(;;)  /* forever */
			{
			/* Le^ a linha enviada pelo cliente
			*/
			recv_sz = recv( client_socket, buffer, sizeof(buffer), 0 );
			if( recv_sz == SOCKET_ERROR )
				{
				if( errno != EPIPE  &&  errno != ECONNRESET )
					perror_and_exit( "recv()" );  /* houve um erro, ou... */
				/* ...o outro lado desligou-se */
				recv_sz = 0;
				}
			if( recv_sz == 0 )
				break;  /* EOF = o cliente desligou */
			buffer[recv_sz] = '\0';  /* assegurar-me que buffer e' uma string em C */

			/* Prepara em "buffer" a resposta ao comando pedido
			*/
			assert( sizeof(buffer) == sizeof(buffer2) );
			if( sscanf(buffer, "diz %[^\r\n]", buffer2) == 1 )
				{
				sprintf( buffer, "DIGO: %s\n", buffer2 );
				}
			else if( sscanf(buffer, "raiz %lf", &num) == 1 )
				{
				sprintf( buffer, "RAIZ QUADRADA DE %lf = %lf\n", num, sqrt(num) );
				}
			else if( sscanf(buffer, "sha1 %d%n", &size, &i) == 1 )
				{
				i++;  /* inclui o '\n' final */
				file_buffer = recv_file( client_socket, size, buffer+i, recv_sz-i );
				if( file_buffer == NULL )
					sprintf( buffer, "SHA-1 DO FICHEIRO NAO PODE SER CALCULADA!\n" );
				else
					{
					assert( sizeof(buffer2) >= 20 );
					sha1( file_buffer, size, (unsigned char*) buffer2 );
					free( file_buffer );
					strcpy( buffer, "SHA-1 DO FICHEIRO = " );
					p = buffer + strlen(buffer);
					for( i = 0;  i < 20;  i++, p += 2 )
						sprintf( p, "%02x", (int) (unsigned char) buffer2[i] );
					strcpy( p, "\n" );
					}
				}
			else if( strncmp(buffer, "desliga", 7) == 0 )
				{
				break;
				}
			else
				{
				strcpy( buffer, TEXT_HELP );
				}

			/* Envia a resposta
			*/
			if( send(client_socket, buffer, strlen(buffer), 0) == INVALID_SOCKET )
				{
				if( errno != EPIPE  &&  errno != ECONNRESET )
					perror_and_exit( "send()" );  /* houve um erro, ou... */
				/* ...o outro lado desligou-se */
				break;
				}
			}

		/* Exibir informacao sobre quem se desligou
		*/
		if( client_sa.sin_family == AF_INET )
			{
			printf( "Desligou-se o cliente no endereco IPv4 %s:%u\n\n",
				inet_ntoa(client_sa.sin_addr), ntohs(client_sa.sin_port) );
			}
		else
			puts( "Desligou-se o cliente de uma familia de enderecos que nao IPv4\n" );

        	/* Fechar o socket do cliente
        	*/
		shutdown( client_socket, 2 );
		}

	/* Nunca chega aqui; mas seria assim:
	*/
       	shutdown( server_socket, 2 );
#ifdef WIN32
	WSACleanup();
#endif
	return EXIT_SUCCESS;
}
